/* Copyright (C) 1999 TcX AB & Monty Program KB & Detron HB

   This software is distributed with NO WARRANTY OF ANY KIND.  No author or
   distributor accepts any responsibility for the consequences of using it, or
   for whether it serves any particular purpose or works at all, unless he or
   she says so in writing.  Refer to the Free Public License (the "License")
   for full details.
   Every copy of this file must include a copy of the License, normally in a
   plain ASCII text file named PUBLIC.	The License grants you the right to
   copy, modify and redistribute this file, but only under certain conditions
   described in the License.  Among other things, the License requires that
   the copyright notice and this notice be preserved on all copies. */

/* Uppdaterare nuvarande record i en pisam-databas */

#include "myisamdef.h"
#ifdef	__WIN32__
#include <errno.h>
#endif

/* Change the current row in the table */

int mi_update(register MI_INFO *info, const byte *oldrec, byte *newrec)
{
  int flag,key_changed,save_errno;
  reg3 my_off_t pos;
  uint i;
  uchar old_key[MI_MAX_KEY_BUFF],*new_key;
  bool auto_key_changed=0;
  ulonglong changed;
  ha_checksum old_checksum;
  DBUG_ENTER("mi_update");

  if (!(info->update & HA_STATE_AKTIV))
  {
    my_errno=HA_ERR_KEY_NOT_FOUND;
    DBUG_RETURN(-1);
  }
  if (info->s->options & HA_OPTION_READ_ONLY_DATA)
  {
    my_errno=EACCES;
    DBUG_RETURN(-1);
  }
  pos=info->lastpos;
#ifndef NO_LOCKING
  if (_mi_readinfo(info,F_WRLCK,1)) DBUG_RETURN(-1);
#endif

  if (info->s->calc_checksum)
    old_checksum=info->checksum=(*info->s->calc_checksum)(info,oldrec);
  if ((*info->s->compare_record)(info,oldrec))
  {
    save_errno=my_errno;
    goto err_end;			/* Record has changed */
  }

  /* Calculate and check all unique constraints */
  for (i=0 ; i < info->s->state.header.uniques ; i++) 
  {
    MI_UNIQUEDEF *def=info->s->uniqueinfo+i;
    if (mi_unique_comp(def, newrec, oldrec,1) &&
	mi_check_unique(info, def, newrec, mi_unique_hash(def, newrec),
			info->lastpos))
      goto err;
  }

  if (_mi_mark_file_changed(info))
    goto err;

  /* Check which keys changed from the original row */

  new_key=info->lastkey+info->s->base.max_key_length;
  key_changed=HA_STATE_KEY_CHANGED;	/* We changed current database */
					/* Remove key that didn't change */
  changed=0;
  for (i=0 ; i < info->s->base.keys ; i++)
  {
    if (((ulonglong) 1 << i) & info->s->state.key_map)
    {
      uint new_length=_mi_make_key(info,i,new_key,newrec,pos);
      uint old_length=_mi_make_key(info,i,old_key,oldrec,pos);
      if (new_length != old_length ||
	  memcmp((byte*) old_key,(byte*) new_key,new_length))
      {
	if ((int) i == info->lastinx)
	  key_changed|=HA_STATE_WRITTEN;	/* Mark that keyfile changed */
	changed|=((ulonglong) 1 << i);
	if (_mi_ck_delete(info,i,old_key,old_length)) goto err;
	if (_mi_ck_write(info,i,new_key,new_length)) goto err;
	if (info->s->base.auto_key == i+1)
	  auto_key_changed=1;
      }
    }
  }

  if (info->s->calc_checksum)
    info->checksum=(*info->s->calc_checksum)(info,newrec);
  if ((*info->s->update_record)(info,pos,newrec))
    goto err;
  if (auto_key_changed)
    update_auto_increment(info,newrec);
  if (info->s->calc_checksum)
    info->s->state.checksum+=(info->checksum - old_checksum);

  info->update= HA_STATE_CHANGED | HA_STATE_AKTIV | key_changed;
  myisam_log_record(MI_LOG_UPDATE,info,newrec,info->lastpos,0);
  VOID(_mi_writeinfo(info,key_changed ?  WRITEINFO_UPDATE_KEYFILE : 0));
  allow_break();				/* Allow SIGHUP & SIGINT */
  DBUG_RETURN(0);

err:
  DBUG_PRINT("error",("key: %d  errno: %d",i,my_errno));
  save_errno=my_errno;
  if (my_errno == HA_ERR_FOUND_DUPP_KEY || my_errno == HA_ERR_RECORD_FILE_FULL)
  {
    info->errkey= (int) i;
    flag=0;
    do
    {
      if (((ulonglong) 1 << i) & changed)
      {
	uint new_length=_mi_make_key(info,i,new_key,newrec,pos);
	uint old_length= _mi_make_key(info,i,old_key,oldrec,pos);
	if ((flag++ && _mi_ck_delete(info,i,new_key,new_length)) ||
	    _mi_ck_write(info,i,old_key,old_length))
	  break;
      }
    } while (i-- != 0);
  }
  else
    info->s->state.changed|=2;			/* mark table crashed */
  info->update= HA_STATE_CHANGED | HA_STATE_AKTIV | key_changed;

 err_end:
  myisam_log_record(MI_LOG_UPDATE,info,newrec,info->lastpos,my_errno);
  VOID(_mi_writeinfo(info,WRITEINFO_UPDATE_KEYFILE));
  allow_break();				/* Allow SIGHUP & SIGINT */
  if (save_errno == HA_ERR_KEY_NOT_FOUND)
    my_errno=HA_ERR_CRASHED;
  else
    my_errno=save_errno;
  DBUG_RETURN(-1);
} /* mi_update */
